import hanabisetting.HanabiSettings;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Panel;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLEncoder;
import java.util.ArrayList;

/**
 * GraphPanel.java Ot\pl킷NX
 * 
 * @author MegumiAraki
 * @date 2007/11/26
 */

class GraphPanel extends Panel
		implements
			Runnable,
			MouseListener,
			MouseMotionListener {
	Graph graph;// \plOt
	int nnodes;// m[h̐
	Node nodes[] = new Node[100];

	int nedges;// ֌W̐
	Edge edges[] = new Edge[200];

	Thread relaxer;
	boolean stress;
	boolean random;
	boolean relatedWordSearch;

	int numMouseButtonsDown = 0;

	/**
	 * RXgN^
	 */
	GraphPanel(Graph graph) {
		this.graph = graph;
		addMouseListener(this);
	}

	/**
	 * m[h
	 */
	int findNode(String lbl, String type) {
		for (int i = 0; i < nnodes; i++) {
			if (nodes[i].lbl.equals(lbl)) {
				return i;// ΁Am[hԍԂ
			}
		}
		return addNode(lbl, type);// Ȃ΁Am[hǉ
	}

	/**
	 * m[hǉ
	 */
	int addNode(String lbl, String type) {
		Node n = new Node();
		n.x = 200;
		n.y = 200;
		n.lbl = lbl;
		n.type = type;
		nodes[nnodes] = n;
		return nnodes++;
	}

	/**
	 * ֌Wǉ
	 */
	void addEdge(String from, String fromType, String to, String toType, int len) {
		Edge e = new Edge();
		e.from = findNode(from, fromType);
		e.to = findNode(to, toType);
		e.len = len;
		edges[nedges++] = e;
	}

	/**
	 * Ot\plJn
	 */
	public void run() {
		Thread me = Thread.currentThread();
		while (relaxer == me) {
			relax();
			if (random && (Math.random() < 0.03)) {
				Node n = nodes[(int) (Math.random() * nnodes)];
				if (!n.fixed) {
					n.x += 100 * Math.random() - 50;
					n.y += 100 * Math.random() - 50;
				}
				graph.play(graph.getCodeBase(), "audio/drip.au");
			}
			try {
				Thread.sleep(100);
			} catch (InterruptedException e) {
				break;
			}
		}
	}

	// \
	synchronized void relax() {
		for (int i = 0; i < nedges; i++) {
			Edge e = edges[i];
			double vx = nodes[e.to].x - nodes[e.from].x;
			double vy = nodes[e.to].y - nodes[e.from].y;
			double len = Math.sqrt(vx * vx + vy * vy);
			len = (len == 0) ? .0001 : len;
			double f = (edges[i].len - len) / (len * 3);
			double dx = f * vx;
			double dy = f * vy;

			nodes[e.to].dx += dx;
			nodes[e.to].dy += dy;
			nodes[e.from].dx += -dx;
			nodes[e.from].dy += -dy;
		}

		for (int i = 0; i < nnodes; i++) {
			Node n1 = nodes[i];
			double dx = 0;
			double dy = 0;

			for (int j = 0; j < nnodes; j++) {
				if (i == j) {
					continue;
				}
				Node n2 = nodes[j];
				double vx = n1.x - n2.x;
				double vy = n1.y - n2.y;
				double len = vx * vx + vy * vy;
				if (len == 0) {
					dx += Math.random();
					dy += Math.random();
				} else if (len < 100 * 100) {
					dx += vx / len;
					dy += vy / len;
				}
			}
			double dlen = dx * dx + dy * dy;
			if (dlen > 0) {
				dlen = Math.sqrt(dlen) / 2;
				n1.dx += dx / dlen;
				n1.dy += dy / dlen;
			}
		}

		Dimension d = getSize();
		for (int i = 0; i < nnodes; i++) {
			Node n = nodes[i];
			if (!n.fixed) {
				n.x += Math.max(-5, Math.min(5, n.dx));
				n.y += Math.max(-5, Math.min(5, n.dy));
			}
			if (n.x < 0) {
				n.x = 0;
			} else if (n.x > d.width) {
				n.x = d.width;
			}
			if (n.y < 0) {
				n.y = 0;
			} else if (n.y > d.height) {
				n.y = d.height;
			}
			n.dx /= 2;
			n.dy /= 2;
		}
		repaint();
	}

	/**
	 * m[h\
	 */
	Node pick;
	boolean pickfixed;
	Image offscreen;
	Dimension offscreensize;
	Graphics offgraphics;

	final Color bgColor = HanabiSettings.GRAPH_BGCOLOR;

	final Color fixedColor = HanabiSettings.KEY_NODE_COLOR;
	final Color selectColor = HanabiSettings.SELECTED_NODE_COLOR;
	final Color edgeColor = HanabiSettings.BASIC_EDGE_COLOR;
	final Color stressColor = HanabiSettings.STRESS_COLOR;
	final Color arcColor1 = HanabiSettings.SHORT_EDGE_COLOR;
	final Color arcColor2 = HanabiSettings.MEDIUM_EDGE_COLOR;
	final Color arcColor3 = HanabiSettings.LONG_EDGE_COLOR;

	final Color type1Color = HanabiSettings.HUMAN_NODE_COLOR;
	final Color type2Color = HanabiSettings.FIELD_NODE_COLOR;

	/**
	 * m[h`悷
	 */
	public void paintNode(Graphics g, Node n, FontMetrics fm) {
		int x = (int) n.x;
		int y = (int) n.y;

		g.setColor((n == pick) ? selectColor : (n.fixed ? fixedColor : (n.type
				.equals(HanabiSettings.TYPE_HUMAN) ? type1Color : type2Color)));
		int w = fm.stringWidth(n.lbl) + 10;
		int h = fm.getHeight() + 4;
		g.fillRect(x - w / 2, y - h / 2, w, h);
		g.setColor(Color.black);
		g.drawRect(x - w / 2, y - h / 2, w - 1, h - 1);
		g.drawString(n.lbl, x - (w - 10) / 2, (y - (h - 4) / 2)
				+ fm.getAscent());
	}

	/**
	 * Ot\plŜ`悷
	 */
	public synchronized void update(Graphics g) {
		Dimension d = getSize();
		if ((offscreen == null) || (d.width != offscreensize.width)
				|| (d.height != offscreensize.height)) {
			offscreen = createImage(d.width, d.height);
			offscreensize = d;
			if (offgraphics != null) {
				offgraphics.dispose();
			}
			offgraphics = offscreen.getGraphics();
			offgraphics.setFont(getFont());
		}

		offgraphics.setColor(bgColor);
		offgraphics.fillRect(0, 0, d.width, d.height);
		for (int i = 0; i < nedges; i++) {
			Edge e = edges[i];
			int x1 = (int) nodes[e.from].x;
			int y1 = (int) nodes[e.from].y;
			int x2 = (int) nodes[e.to].x;
			int y2 = (int) nodes[e.to].y;
			int len = (int) Math.abs(Math.sqrt((x1 - x2) * (x1 - x2)
					+ (y1 - y2) * (y1 - y2))
					- e.len);
			offgraphics.setColor((len < 10) ? arcColor1 : (len < 20
					? arcColor2
					: arcColor3));
			offgraphics.drawLine(x1, y1, x2, y2);
			if (stress) {
				String lbl = String.valueOf(len);
				offgraphics.setColor(stressColor);
				offgraphics.drawString(lbl, x1 + (x2 - x1) / 2, y1 + (y2 - y1)
						/ 2);
				offgraphics.setColor(edgeColor);
			}
		}

		FontMetrics fm = offgraphics.getFontMetrics();
		for (int i = 0; i < nnodes; i++) {
			paintNode(offgraphics, nodes[i], fm);
		}
		g.drawImage(offscreen, 0, 0, null);
	}

	/**
	 * }EXNbNꂽƂ̏
	 */
	public void mouseClicked(MouseEvent e) {
	}

	/**
	 * }EX{^ꂽƂ̏
	 */
	public void mousePressed(MouseEvent e) {
		numMouseButtonsDown++;
		addMouseMotionListener(this);
		double bestdist = Double.MAX_VALUE;

		int x = e.getX();
		int y = e.getY();
		for (int i = 0; i < nnodes; i++) {
			Node n = nodes[i];
			double dist = (n.x - x) * (n.x - x) + (n.y - y) * (n.y - y);
			if (dist < bestdist) {
				pick = n;
				bestdist = dist;
			}
		}
		pickfixed = pick.fixed;
		pick.fixed = true;
		pick.x = x;
		pick.y = y;

		// NbNꂽ猟ʂURLɔ
		String nodeWord = pick.lbl;// 
		ArrayList<String> relatedWords = getRelatedWords(nodeWord);
		String relatedWordForSearch = RelatedWordsToString(relatedWords);// ֘A

		try {
			nodeWord = URLEncoder.encode(nodeWord, "UTF-8");
		} catch (UnsupportedEncodingException e2) {
			e2.printStackTrace();
		}
		URL newURL = null;

		if (relatedWordSearch) {
			try {
				newURL = new URL(HanabiSettings.NODE_CLICK_SEARCH_URL
						+ nodeWord + relatedWordForSearch
						+ HanabiSettings.NODE_CLICK_SEARCH_ENDOCING
						+ HanabiSettings.NODE_CLICK_SEARCH_DOMAIN);
			} catch (MalformedURLException e1) {
				e1.printStackTrace();
			}
		} else {
			String searchHint = "";
			try {
				newURL = new URL(HanabiSettings.NODE_CLICK_SEARCH_URL
						+ nodeWord + "+"
						+ URLEncoder.encode(searchHint, "UTF-8")
						+ HanabiSettings.NODE_CLICK_SEARCH_ENDOCING
						+ HanabiSettings.NODE_CLICK_SEARCH_DOMAIN);
			} catch (MalformedURLException e1) {
				e1.printStackTrace();
			} catch (UnsupportedEncodingException e2) {
				e2.printStackTrace();
			}
		}
		graph.getAppletContext().showDocument(newURL, "search");// Et[Ɍʃy[W\

		repaint();
		e.consume();
	}
	/**
	 * }EX{^ꂽƂ̏
	 */
	public void mouseReleased(MouseEvent e) {
		numMouseButtonsDown--;
		removeMouseMotionListener(this);

		pick.fixed = pickfixed;
		pick.x = e.getX();
		pick.y = e.getY();
		if (numMouseButtonsDown == 0) {
			pick = null;
		}

		repaint();
		e.consume();
	}

	/**
	 * }EXƂ̏
	 */
	public void mouseEntered(MouseEvent e) {
	}

	/**
	 * }EXoƂ̏
	 */
	public void mouseExited(MouseEvent e) {
	}

	/**
	 * }EXhbOꂽƂ̏
	 */
	public void mouseDragged(MouseEvent e) {
		pick.x = e.getX();
		pick.y = e.getY();
		repaint();
		e.consume();
	}

	/**
	 * }EXƂ̏
	 */
	public void mouseMoved(MouseEvent e) {
	}

	/**
	 * Ot\plJn
	 */
	public void start() {
		relaxer = new Thread(this);
		relaxer.start();
	}

	/**
	 * Ot\plI
	 */
	public void stop() {
		relaxer = null;
	}

	/**
	 * ֘AĂ擾
	 */
	private ArrayList<String> getRelatedWords(String nodeWord) {
		ArrayList<String> relatedWords = new ArrayList<String>();
		for (int i = 0; i < nedges; i++) {
			Edge edge = edges[i];
			String fromNodeLabel = nodes[edge.from].lbl;
			String toNodeLabel = nodes[edge.to].lbl;

			if (fromNodeLabel.equals(nodeWord)) {
				relatedWords.add(toNodeLabel);
			} else if (toNodeLabel.equals(nodeWord)) {
				relatedWords.add(fromNodeLabel);
			}
		}
		return relatedWords;
	}

	/**
	 * ֘AĂpɕϊ
	 */
	private String RelatedWordsToString(ArrayList<String> relatedWords) {
		String urlText = "";

		for (String relatedWord : relatedWords) {
			try {
				urlText = urlText + "+"
						+ URLEncoder.encode(relatedWord, "UTF-8");
			} catch (UnsupportedEncodingException e) {
				e.printStackTrace();
			}
		}
		return urlText;
	}
}
